import os
import sys
import argparse
import logging
import importlib

from .client import ChannelsDiscordClient

logger = logging.getLogger(__name__)


class CLI(object):
    """
    Primary command-line interface for Discord -> channels bridge
    """
    description = "Django Channels <-> Discord Interface server"

    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description=self.description,
        )
        self.parser.add_argument(
            '-a',
            '--application',
            dest='application',
            help='The django channels application as path.to.module:instance.path',
            default=os.environ.get('CHANNELS_DISCORD_APPLICATION', None),
        )
        self.parser.add_argument(
            '-t',
            '--token',
            dest='token',
            help='OAuth or Bot token used to connect to discord',
            default=os.environ.get('DISCORD_BOT_TOKEN', None),
        )
        self.parser.add_argument(
            '-v',
            '--verbosity',
            type=int,
            help='How verbose to make the output. Default is 1',
            default=1,
        )

    @classmethod
    def entrypoint(cls):
        """
        Many entrypoint for external starts
        """
        cls().run(sys.argv[1:])

    def run(self, args):
        """
        Mounts the Discord interface server based on the raw arguments passed in
        """
        args = self.parser.parse_args(args)

        # Set up logging
        logging.basicConfig(
            level={
                0: logging.WARN,
                1: logging.INFO,
                2: logging.DEBUG,
            }[args.verbosity],
            format="%(asctime)-15s %(levelname)-8s %(message)s",
        )

        # suppress logs from imported libraries to desired level
        if args.verbosity < 2:
            disable = {
                0: logging.INFO,
                1: logging.DEBUG,
            }[args.verbosity]
            logging.disable(disable)

        # import the channel layer
        asgi_module, application_path = args.application.split(':', 1)
        application_module = importlib.import_module(asgi_module)
        for part in application_path.split('.'):
            application = getattr(application_module, part)

        client = ChannelsDiscordClient(application)
        client.run(args.token)



import asyncio
import logging

import requests
from discord import Client
from discord.errors import LoginFailure

logger = logging.getLogger(__name__)

DISCORD_API = 'https://discordapp.com/api/v6'


class ChannelsDiscordClient(Client):
    def __init__(self, application, **kwargs):
        self.application = application
        super().__init__(**kwargs)

        self.create_application()

    def create_application(self):
        """
        Handles creating the ASGI application and instatiating the
        send Queue
        """
        self.application_queue = asyncio.Queue()
        application_instance = self.application(
            scope={'type': 'discord'},
            receive=self.application_queue.get,
            send=self.from_consumer
        )
        self.application_instance = asyncio.ensure_future(
            application_instance, loop=self.loop
        )

    def to_application(self, message):
        """
        Handles low-level job of sending info to the consumer
        """
        return self.application_queue.put_nowait(message)

    async def on_ready(self):
        """
        Sends the 'ready' message to the consumer.  Useful for running
        tasks upon initially connecting to discord
        """
        self.to_application({
            'type': 'discord.ready',
        })

    async def from_consumer(self, message):
        """
        Receives message from channels from the consumer.  Message should have the format:
            {
                'type': 'discord.send',
                'action': VALID_ACTION_TYPE
            }
        """
        if "type" not in message:
            raise ValueError("Message has no type defined")

        elif message['type'] == 'discord.send':
            action_type = message.get('action', '').lower()

            handler = getattr(self, '_handle_{}'.format(action_type))
            await handler(message)

        else:
            raise ValueError("Cannot handle message type %s!" % message["type"])

    async def _handle_dm(self, message):
        """
        Sends a direct message to a user, based on user id
        """
        user_id = message.get('user_id')
        text = message.get('text')

        assert user_id is not None and text is not None, (
            "Sending a DM requires both `user_id` and `text` keys"
        )

        # try fetching from local cache
        user = self.get_user(user_id)

        # If user info is not cached, try getting from the API
        if user is None:
            user = await self.fetch_user(user_id)

        if user is not None:
            await user.send(content=text)

    async def _handle_send_to_channel(self, message):
        """
        Sends a direct message to a channel, based on channel id
        """
        channel_id = message.get('channel_id')
        text = message.get('text')

        assert channel_id is not None and text is not None, (
            "Sending a DM requires both `channel_id` and `text` keys"
        )

        # try fetching from local cache
        channel = self.get_channel(channel_id)

        # if channel is not cached, try getting from the API
        if channel is None:
            channel = await self.fetch_channel(channel_id)

        if channel is not None:
            await channel.send(content=text)



import logging

from channels.consumer import AsyncConsumer
from channels.exceptions import InvalidChannelLayerError, StopConsumer

logger = logging.getLogger(__name__)


class DiscordConsumer(AsyncConsumer):
    """
    Base Discord Consumer; Implements basic hooks for interacting with
    the Discord Interface Server
    """
    async def send_action(self, action, **kwargs):
        await self.send({
            'type': 'discord.send',
            'action': action,
            **kwargs,
        })

    async def discord_ready(self, event):
        """
        Runs when initially connected to discord.  This method is primarily
        used for internal tasks, such as adding groups to the channel layer.
        Any tasks meant to be run on initially connecting should be called
        by overriding the `ready` method in your app's consumer
        """
        try:
            for group in self.groups:
                await self.channel_layer.group_add(group, self.channel_name)
        except AttributeError:
            raise InvalidChannelLayerError("BACKEND is unconfigured or doesn't support groups")
        await self.ready()

    async def ready(self):
        """
        Hook for any action(s) to be run on connecting to Discord server
        """
        pass
